/*************************************************************************
 *
 *  Game of Life -- a screen saver for Windows
 *  Copyright (C) 2007, 2013 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  cycle.c: main animation core
 *
 ************************************************************************/


#include "life.h"
#include "out_win.h"


WORD ix, iy, vx, vy;
WORD nel, ner; /* new edge_* */
LPBYTE stn, stns, sup, sups, sdn, sdns;
LPBYTE stt, stts, tup, tups, tdn, tdns;
WORD stny, supy, sdny;
DWORD d;
WORD h;


#ifdef LIFE_DEBUGMODE
extern VOID debug_neighbour_counter(WORD, WORD, LPBYTE);
#endif


#define HEALTH_CHECK(OUCH) {											\
	sups = state_next;													\
	stns = sups + UNIVERSE_W;											\
	sdns = stns + UNIVERSE_W;											\
	for (vy = 1; vy <= ACTIVE_H; vy++) {								\
		for (sup = sups, stn = stns, sdn = sdns,						\
			vx = 1; vx <= ACTIVE_W; vx++) {								\
			h  = ((*sup) & 0x10) + (sup[1] & 0x10) + (sup[2] & 0x10);	\
			h += ((*stn) & 0x10) +                   (stn[2] & 0x10);	\
			h += ((*sdn) & 0x10) + (sdn[1] & 0x10) + (sdn[2] & 0x10);	\
			h >>= 4;													\
			sup++; stn++; sdn++;										\
			if (((*stn) & 0x0F) != h)	/* does this hurt? */			\
				OUCH;													\
		}																\
		sups += UNIVERSE_W;												\
		stns += UNIVERSE_W;												\
		sdns += UNIVERSE_W;												\
	}																	\
}


/* mutate_edges()

	creates live and dead cells on
	  the edges of the universe */

VOID mutate_edges(VOID) {

	#if 1

	BYTE new_state, new_colour;

	#define EDGE_NEIGHBOURS  3

	d = 0;
	new_colour = 0x10 | EDGE_NEIGHBOURS;

	#define mutate_left_edge {												\
		edge_L[supy] = edge_L[stny] = edge_L[sdny] = 1;						\
		if (edge_R[supy] < 2) edge_R[supy] = 2;								\
		if (edge_R[stny] < 2) edge_R[stny] = 2;								\
		if (edge_R[sdny] < 2) edge_R[sdny] = 2;								\
	}
	#define mutate_right_edge {												\
		edge_R[supy] = edge_R[stny] = edge_R[sdny] = ner;					\
		if (!edge_L[supy]) edge_L[supy] = nel;								\
		if (!edge_L[stny]) edge_L[stny] = nel;								\
		if (!edge_L[sdny]) edge_L[sdny] = nel;								\
	}
	#define mutate_top_or_bottom_edge {										\
		ner = vx;															\
		if (!nel) nel = vx;													\
	}

	#define do_mutation(edgefunc) {											\
		if (!d) {															\
			rnd();															\
			d = 0xC0000000;													\
		}																	\
		new_state = (random_number & d) ? 0x00 : 0x10;						\
		if (((*stt) & 0x10) == new_state) {									\
			*stn = ((*stt) & 0x30) | EDGE_NEIGHBOURS;						\
			new_colour ^= 0x20;												\
			d >>= 2;														\
			continue;														\
		}																	\
		edgefunc;															\
		if (new_state) {													\
			(*sups)++; (*stns)++; (*sdns)++;								\
			*stn = new_colour;												\
		}																	\
		else {																\
			(*sups)--; (*stns)--; (*sdns)--;								\
			*stn = EDGE_NEIGHBOURS;											\
		}																	\
		d >>= 2;															\
	}

	#define mutate_vertical_edge(EDGEX, SIDESTEP, edgefunc) {				\
		stt  = state_this + (EDGEX) + UNIVERSE_W;							\
		stn  = state_next + (EDGEX);										\
		sups = stn  + (SIDESTEP);											\
		stn += UNIVERSE_W;													\
		stns = stn  + (SIDESTEP);											\
		sdns = stns + UNIVERSE_W;											\
		supy = 0;															\
		stny = 1;															\
		sdny = 2;															\
																			\
		for (iy = ACTIVE_H;													\
			iy;																\
			iy--,															\
			supy++, stny++, sdny++,											\
			stt += UNIVERSE_W, stn += UNIVERSE_W,							\
			sups = stns, stns = sdns, sdns += UNIVERSE_W)					\
																			\
			do_mutation(edgefunc)											\
	}

	#define mutate_horizontal_edge(EDGEY, SIDESTEP, SIDEY, edgefunc) {		\
		stt  = state_this + (EDGEY) + 1;									\
		stn  = state_next + (EDGEY) + 1;									\
		stns = stn  + (SIDESTEP);											\
		sups = stns - 1;													\
		sdns = stns + 1;													\
		stny = SIDEY;														\
		nel  = ner = 0;														\
																			\
		for (ix = ACTIVE_W, vx = 1;											\
			ix;																\
			ix--, vx++,														\
			stt++, stn++,													\
			sups++, stns++, sdns++)											\
																			\
			do_mutation(edgefunc)											\
																			\
		if (nel) {															\
			nel--;															\
			if (!nel)														\
				nel = 1;													\
			ix = edge_L[stny];												\
			if ((!ix) || (ix > nel))										\
				edge_L[stny] = nel;											\
			ner++; ner++;													\
			if (ner >= UNIVERSE_W)											\
				ner = UNIVERSE_W - 1;										\
			if (edge_R[stny] < ner)											\
				edge_R[stny] = ner;											\
		}																	\
	}

	#ifdef LIFE_DEBUGMODE
	mutated = 1;
	#endif

	nel = UNIVERSE_W - 2; ner = UNIVERSE_W - 1;

	mutate_vertical_edge(  0,  1, mutate_left_edge );				/* left   */
	mutate_vertical_edge(ner, -1, mutate_right_edge);				/* right  */
	mutate_horizontal_edge(0,
		UNIVERSE_W,              1, mutate_top_or_bottom_edge);		/* top    */
	mutate_horizontal_edge(UNIVERSE_W_MULTIPLY(UNIVERSE_H - 1),
		-UNIVERSE_W, UNIVERSE_H - 2, mutate_top_or_bottom_edge);	/* bottom */

	state_next[                         0] = EDGE_NEIGHBOURS;
	state_next[UNIVERSE_W    -          1] = EDGE_NEIGHBOURS;
	state_next[UNIVERSE_SIZE - UNIVERSE_W] = EDGE_NEIGHBOURS;
	state_next[UNIVERSE_SIZE -          1] = EDGE_NEIGHBOURS;

	#endif
}


/* cycle()

	handles the evolution from one generation
	  to the next and redraws the screen */

CHAR cycle(VOID) {

	DWORD dwT, dwS;
	BYTE shapeT, shapeB, shapeD;
	CHAR return_value = 0;
	#ifdef LIFE_DEBUGMODE
	TCHAR szDebug[11];
	#endif

	draw_func_header

	#ifdef LIFE_DEBUGMODE
	mutated = 0;
	#endif

	#define skip_empty_row { if (!ix) continue; }

	#define setup_x_loop(empty_row_action) {    \
		ix   = edge_r[iy] = edge_R[iy];         \
		edge_l[iy] = 0;                         \
		empty_row_action;                       \
		left = edge_l[iy] = edge_L[iy];         \
		ix -= left;                             \
		edge_L[iy] = edge_R[iy] = 0;            \
		st = sts + left;                        \
		sn = sns + left;                        \
	}

	/* copy top row to state_next */

	for (st = state_this, sn = state_next, ix = UNIVERSE_W; ix; ix--, st++, sn++)
		*sn = *st;

	/* handle evolution from state_this to state_next */

	for (iy = edge_t = 1, edge_b = 0,
		sty = ACTIVE_H;
		sty;
		sty--, iy++,
		sts += UNIVERSE_W, sns += UNIVERSE_W) {

		/* copy left and right columns to state_next */

		sns[             0] = sts[             0];
		sns[UNIVERSE_W - 1] = sts[UNIVERSE_W - 1];

		setup_x_loop(skip_empty_row);

		for ( ; ix; ix--) {
			e = *st;
			do_evolution_of_one_cell(select_colour_of_new_cell);
		}
	}

	/* copy bottom row to state_next */

	for (st = sts, sn = sns, ix = UNIVERSE_W; ix; ix--, st++, sn++)
		*sn = *st;

	if (!edge_b)
		goto next_generation;

	/* update neighbour counters */

	edge_b++;
	d = UNIVERSE_W_MULTIPLY(edge_t);

	stns = state_next + d;
	stts = state_this + d;

	stny = supy = sdny = edge_t; supy--; sdny++;
	sups = stns - UNIVERSE_W;
	sdns = stns + UNIVERSE_W;
	tups = stts - UNIVERSE_W;
	tdns = stts + UNIVERSE_W;

	#define continue_x_cycle_loop {                                         \
		iy--;                                                               \
		supy++; sups  = stns;                                               \
		stny++; stns  = sdns;                                               \
		sdny++; sdns += UNIVERSE_W;                                         \
	}

	#define start_x_cycle_loop(skip_action) {                               \
		ner = edge_r[stny];                                                 \
		if (!ner) {                                                         \
			continue_x_cycle_loop;                                          \
			skip_action;                                                    \
		}                                                                   \
		nel = vx = ix = edge_l[stny];                                       \
		stn = stns + vx;                                                    \
		vx--;                                                               \
		/*stt = stts + vx;*/                                                \
		sup = sups + vx; sdn = sdns + vx;                                   \
		tup = tups + vx; tdn = tdns + vx;                                   \
		vx  = ner - ix;                                                     \
		nel = ner = 0;                                                      \
	}

	#define end_x_cycle_loop {                                              \
		if (nel) {                                                          \
			nel--;                                                          \
			if (!nel)                                                       \
				nel = 1;                                                    \
			ix = edge_L[supy];                                              \
			if ((!ix) || (ix > nel))                                        \
				edge_L[supy] = nel;                                         \
			ix = edge_L[stny];                                              \
			if ((!ix) || (ix > nel))                                        \
				edge_L[stny] = nel;                                         \
			ix = edge_L[sdny];                                              \
			if ((!ix) || (ix > nel))                                        \
				edge_L[sdny] = nel;                                         \
			ner++; ner++;                                                   \
			if (ner >= UNIVERSE_W)                                          \
				ner = UNIVERSE_W - 1;                                       \
			if (edge_R[supy] < ner)                                         \
				edge_R[supy] = ner;                                         \
			if (edge_R[stny] < ner)                                         \
				edge_R[stny] = ner;                                         \
			if (edge_R[sdny] < ner)                                         \
				edge_R[sdny] = ner;                                         \
		}                                                                   \
		continue_x_cycle_loop;                                              \
	}

	#define cycle_one_cell(change_action, default_action) {                 \
		e = (*stn) & 0xB0;                                                  \
		switch (e) {                                                        \
			case 0x90: case 0xB0: /* births */                              \
				(*stn) &= /*0x9F;*/0x3F;                                    \
				change_action;                                              \
				stn--; (*stn)++;                                            \
				stn++;                                                      \
				stn++; (*stn)++;                                            \
				       (*sup)++;                                            \
				sup++; (*sup)++;                                            \
				sup++; (*sup)++;                                            \
				sup--;                                                      \
				       (*sdn)++;                                            \
				sdn++; (*sdn)++;                                            \
				sdn++; (*sdn)++;                                            \
				sdn--;                                                      \
				ner = ix; if (!nel) nel = ix;                               \
				break;                                                      \
			case 0x80: case 0xA0: /* deaths */                              \
				(*stn) &= 0x0F;                                             \
				change_action;                                              \
				stn--; (*stn)--; stn++; stn++; (*stn)--;                    \
				(*sup)--; sup++; (*sup)--; sup++; (*sup)--; sup--;          \
				(*sdn)--; sdn++; (*sdn)--; sdn++; (*sdn)--; sdn--;          \
				ner = ix; if (!nel) nel = ix;                               \
				break;                                                      \
			default:                                                        \
				default_action;                                             \
				stn++; sup++; sdn++;                                        \
		}                                                                   \
		ix++; vx--;                                                         \
	}

	#define update_shapeT {                                                 \
		shapeT <<= 2; shapeT |= (sup[1] & 0x30) | ((sup[2] & 0x30) >> 2);   \
	}
	#define update_shapeB {                                                 \
		shapeB <<= 2; shapeB |= (stn[0] & 0x30) | ((stn[1] & 0x30) >> 2);   \
	}
	#define update_shapeD {                                                 \
		shapeD <<= 2; shapeD |= (sdn[1] & 0x30) | ((sdn[2] & 0x30) >> 2);   \
	}
	#define update_shape {                                                  \
		update_shapeT; update_shapeB; update_shapeD;                        \
	}
	#define update_shapeT_and_B {                                           \
		update_shapeT; update_shapeB;                                       \
	}
	#define update_shapeB_and_D {                                           \
		update_shapeB; update_shapeD;                                       \
	}

	#define add_tile(OFFSET, ROW, COL)                                      \
		add_tile_to_redraw_queue(dwT + OFFSET,                              \
			wTileOffset[ROW],                                               \
			wTileOffset[COL])

	#define draw_tile_full {                                                \
		update_shape;                                                       \
		add_tile(            0,  shapeB >> 4       ,  shapeT >> 4       );  \
		add_tile(            1, (shapeB >> 2) & 0xF, (shapeT >> 2) & 0xF);  \
		add_tile(VISIBLE_W    ,  shapeD >> 4       ,  shapeB >> 4       );  \
		add_tile(VISIBLE_W + 1, (shapeD >> 2) & 0xF, (shapeB >> 2) & 0xF);  \
	}
	#define draw_tile_right {                                               \
		update_shape;                                                       \
		add_tile(            0, (shapeB >> 2) & 0xF, (shapeT >> 2) & 0xF);  \
		add_tile(VISIBLE_W    , (shapeD >> 2) & 0xF, (shapeB >> 2) & 0xF);  \
	}
	#define draw_tile_left {                                                \
		update_shape;                                                       \
		add_tile(            0,  shapeB >> 4       ,  shapeT >> 4       );  \
		add_tile(VISIBLE_W    ,  shapeD >> 4       ,  shapeB >> 4       );  \
	}
	#define draw_tile_bottom {                                              \
		update_shapeB_and_D;                                                \
		add_tile(            0,  shapeD >> 4       ,  shapeB >> 4       );  \
		add_tile(            1, (shapeD >> 2) & 0xF, (shapeB >> 2) & 0xF);  \
	}
	#define draw_tile_top {                                                 \
		update_shapeT_and_B;                                                \
		add_tile(            0,  shapeB >> 4       ,  shapeT >> 4       );  \
		add_tile(            1, (shapeB >> 2) & 0xF, (shapeT >> 2) & 0xF);  \
	}
	#define draw_tile_b_right {                                             \
		update_shapeB_and_D;                                                \
		add_tile(            0, (shapeD >> 2) & 0xF, (shapeB >> 2) & 0xF);  \
	}
	#define draw_tile_b_left {                                              \
		update_shapeB_and_D;                                                \
		add_tile(            0,  shapeD >> 4       ,  shapeB >> 4       );  \
	}
	#define draw_tile_t_right {                                             \
		update_shapeT_and_B;                                                \
		add_tile(            0, (shapeB >> 2) & 0xF, (shapeT >> 2) & 0xF);  \
	}
	#define draw_tile_t_left {                                              \
		update_shapeT_and_B;                                                \
		add_tile(            0,  shapeB >> 4       ,  shapeT >> 4       );  \
	}

	if (edge_t >= WIN_VISIBLE_BOTTOM)
		goto bottom_x_cycle_loop_done;

	if (edge_t <= MARGIN_T) {

		/* handle cells above visible area */

		iy = (edge_b < MARGIN_T) ? edge_b : MARGIN_T;

		for (iy -= edge_t; iy; ) {

			start_x_cycle_loop(continue);

			for ( ; vx; )
				cycle_one_cell( { }, { } );

			end_x_cycle_loop;
		}

		/* handle cells on top visible row */

		start_x_cycle_loop(goto top_x_cycle_loop_done);

		stn--;
		shapeB = (*stn) & 0x30; stn++;
		shapeD = (*sdn) & 0x30;

		if (ix < MARGIN_L) {

			d = MARGIN_L - ix;
			if (d > vx) d = vx;

			for ( ; d; d--)
				cycle_one_cell(update_shapeB_and_D, update_shapeB_and_D);

			if (!vx)
				goto end_top_x_cycle_loop;
		}

		dwT = 0;
		if (ix == MARGIN_L) {
			cycle_one_cell(draw_tile_b_right, update_shapeB_and_D);
		}
		else
			dwT = ix - MARGIN_L - 1;

		if (ix <= WIN_VISIBLE_RIGHT) {

			d = WIN_VISIBLE_RIGHT - ix;
			if (d > vx) d = vx;

			for ( ; d; d--, dwT++)
				cycle_one_cell(draw_tile_bottom, update_shapeB_and_D);

			if (vx)
				cycle_one_cell(draw_tile_b_left, { } );
		}

		for ( ; vx; )
			cycle_one_cell( { }, { } );

		end_top_x_cycle_loop:

		end_x_cycle_loop;
	}

	top_x_cycle_loop_done:

	if (stny >= edge_b)
		goto all_updating_done;

	dwS = supy - MARGIN_T;
	dwS *= VISIBLE_W;

	/* handle cells in visible area */

	iy = (edge_b <= WIN_VISIBLE_BOTTOM) ? edge_b : WIN_VISIBLE_BOTTOM;

	for (iy -= stny; iy; dwS += VISIBLE_W) {

		start_x_cycle_loop(continue);

		#ifdef LIFE_DEBUGMODE
		#if 0
		#define catch_the_big_one_when_it_falls(ARGH) {	\
			return_value = -(100 + ARGH);				\
			wsprintf(szDebug, "%1u (%2u..%2u)",			\
				ARGH, edge_t, edge_b);					\
			TextOut(hdcOutput, 467, 0, szDebug, 10);	\
			goto switch_state;							\
		}
		if (sup == NULL) catch_the_big_one_when_it_falls(0);
		if (sup < state_next) catch_the_big_one_when_it_falls(1);
		if (sup >= state_next + UNIVERSE_SIZE) catch_the_big_one_when_it_falls(2);
		#endif
		#endif

		shapeT = (*sup) & 0x30; stn--;
		shapeB = (*stn) & 0x30; stn++;
		shapeD = (*sdn) & 0x30;

		if (ix < MARGIN_L) {

			d = MARGIN_L - ix;
			if (d > vx) d = vx;

			for ( ; d; d--)
				cycle_one_cell(update_shape, update_shape);

			if (!vx)
				goto end_visible_x_cycle_loop;
		}

		dwT = dwS;
		if (ix == MARGIN_L) {
			cycle_one_cell(draw_tile_right, update_shape);
		}
		else
			dwT += ix - MARGIN_L - 1;

		if (ix <= WIN_VISIBLE_RIGHT) {

			d = WIN_VISIBLE_RIGHT - ix;
			if (d > vx) d = vx;

			for ( ; d; d--, dwT++)
				cycle_one_cell(draw_tile_full, update_shape);

			if (vx)
				cycle_one_cell(draw_tile_left, { } );
		}

		for ( ; vx; )
			cycle_one_cell( { }, { } );

		end_visible_x_cycle_loop:

		end_x_cycle_loop;
	}

	visible_x_cycle_loop_done:

	if (stny < edge_b) {

		/* handle cells on bottom visible row */

		start_x_cycle_loop(goto bottom_x_cycle_loop_done);

		shapeT = (*sup) & 0x30; stn--;
		shapeB = (*stn) & 0x30; stn++;

		if (ix < MARGIN_L) {

			d = MARGIN_L - ix;
			if (d > vx) d = vx;

			for ( ; d; d--)
				cycle_one_cell(update_shapeT_and_B, update_shapeT_and_B);

			if (!vx)
				goto end_bottom_x_cycle_loop;
		}

		dwT = WIN_REDRAW_BOTTOM_LEFT;
		if (ix == MARGIN_L) {
			cycle_one_cell(draw_tile_t_right, update_shapeT_and_B);
		}
		else
			dwT += ix - MARGIN_L - 1;

		if (ix <= WIN_VISIBLE_RIGHT) {

			d = WIN_VISIBLE_RIGHT - ix;
			if (d > vx) d = vx;

			for ( ; d; d--, dwT++)
				cycle_one_cell(draw_tile_top, update_shapeT_and_B);

			if (vx)
				cycle_one_cell(draw_tile_t_left, { } );
		}

		for ( ; vx; )
			cycle_one_cell( { }, { } );

		end_bottom_x_cycle_loop:

		end_x_cycle_loop;

		bottom_x_cycle_loop_done:

		/* handle cells below visible area */

		for (iy = edge_b - stny; iy; ) {

			start_x_cycle_loop(continue);

			for ( ; vx; )
				cycle_one_cell( { }, { } );

			end_x_cycle_loop;
		}
	}

	/* draw updated state */

	all_updating_done:

	redraw();

	next_generation:

	generation++;

	if (changing_cells < MUTATION_THRESHOLD)
		mutate_edges();

	#ifdef LIFE_DEBUGMODE

	#if 0				/**** it's a crash test, dummy ****/
	if ((random_number & 0xC000) == 0x8000) {
		vx = random_number % (VISIBLE_W - 1);
		vy = (random_number >> 16) % (VISIBLE_H - 1);
		vx += MARGIN_L + 1;
		vy += MARGIN_T + 1;
		state_next[UNIVERSE_W_MULTIPLY(vy) + vx] ^= 0x10;
	}
	#endif

	#if 1

	#if 1
	for (vy = 0; vy < UNIVERSE_H; vy++) {
		if (!edge_L[vy]) {
			if (edge_R[vy]) {
				return_value = -1;
				break;
			}
		}
		else
			if (edge_R[vy] >= UNIVERSE_W) {
				return_value = -1;
				break;
			}
	}
	#endif

	HEALTH_CHECK( {
		if (vx >= MARGIN_L)
			if (vy >= MARGIN_T) {
				ix = vx - MARGIN_L;
				iy = vy - MARGIN_T;
				if (ix <= VISIBLE_W)
					if (iy <= VISIBLE_H)
						debug_neighbour_counter(ix, iy, stn);
			}
		return_value = -1;
	} );

	#endif

	#if 0
	if (return_value < 0) {

		#define check_edge(XPOS, LETTER) {							\
			if ((*stn) & 0x10) {									\
				e = wsprintf(szDebug, TEXT("%c%u"), LETTER, d);		\
				TextOut(hdcOutput, XPOS, 0, szDebug, e);			\
			}														\
		}

		for (stn = state_next, d = 0;
			d < UNIVERSE_W; d++, stn++)
			check_edge(484, 'T');
		for (stn = state_next + UNIVERSE_SIZE - UNIVERSE_W,
			d = 0; d < UNIVERSE_W; d++, stn++)
			check_edge(540, 'B');
		for (stn = state_next + UNIVERSE_W,
			d = 1; d <=  ACTIVE_H; d++, stn += UNIVERSE_W)
			check_edge(456, 'L');
		for (stn = state_next + (UNIVERSE_W << 1) - 1,
			d = 1; d <=  ACTIVE_H; d++, stn += UNIVERSE_W)
			check_edge(512, 'R');
	}
	#endif

	#endif /* LIFE_DEBUGMODE */

	/* switch state pointers */

	switch_state:

	if (state_flag) {
		state_this = state0;
		state_next = state1;
		state_flag = 0;
	}
	else {
		state_this = state1;
		state_next = state0;
		state_flag = 1;
	}

	return return_value;
}


/* health_check()

	checks the neighbour count of every cell in
	  the active area;
	if there is a count that doesn't match what
	  is really there, nonzero is returned */

CHAR health_check(VOID) {

	HEALTH_CHECK( { return -1; } );

	return 0;
}

