/*************************************************************************
 *
 *  Game of Life -- a screen saver for Windows
 *  Copyright (C) 2007, 2013 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  life.h: main header file
 *
 ************************************************************************/


#ifndef LIFE_H_INCLUDED
#define LIFE_H_INCLUDED


#include <windows.h>


#if 1
#define MARGIN_L                2	/* >= 1 */
#define MARGIN_R                3	/* >= 1 */
#define MARGIN_T                2	/* >= 1 */
#define MARGIN_B                3	/* >= 1 */
#else
#define MARGIN_L               20	/* >= 1 */
#define MARGIN_R               21	/* >= 1 */
#define MARGIN_T               20	/* >= 1 */
#define MARGIN_B               21	/* >= 1 */
#endif

#define MARGIN_T_MINUS_ONE      (MARGIN_T - 1)	/* >= 0 */
#define MARGIN_B_MINUS_ONE      (MARGIN_B - 1)	/* >= 0 */

WORD VISIBLE_W;					/* visible grid width (cells) */
WORD VISIBLE_H;					/* visible grid height (cells) */
WORD UNIVERSE_W;				/* VISIBLE_W + MARGIN_L + MARGIN_R */
WORD UNIVERSE_H;				/* VISIBLE_H + MARGIN_T + MARGIN_B */
WORD ACTIVE_W;					/* UNIVERSE_W - 2 */
WORD ACTIVE_H;					/* UNIVERSE_H - 2 */
DWORD UNIVERSE_SIZE;			/* UNIVERSE_W_MULTIPLY(UNIVERSE_H) */

#define UNIVERSE_W_MULTIPLY(X) ((X) * UNIVERSE_W)

DWORD WIN_VISIBLE_TOP_LEFT;		/* UNIVERSE_W_MULTIPLY(MARGIN_T) + MARGIN_L */
WORD WIN_VISIBLE_RIGHT;			/* VISIBLE_W + MARGIN_L + 1 */
WORD WIN_VISIBLE_BOTTOM;		/* VISIBLE_H + MARGIN_T + 1 */
DWORD WIN_REDRAW_BOTTOM_LEFT;	/* VISIBLE_W * (VISIBLE_H - 1) */


LPBYTE state0, state1;
LPBYTE state_this, state_next;
BYTE state_flag;

/*	state:
		CxVANNNN
		| ||+--+-- 0x0F: live neighbours (0..8)
		| |+------ 0x10: alive
		| +------- 0x20: which one of two colours
		+--------- 0x80: state changing
*/

BYTE evolution[64];

DWORD generation;
DWORD population[2];
DWORD changing_cells;	/* the sum of births and deaths, not counting blinkers */


/* new cells are spawned on the edges of the universe
     whenever changing_cells is less than MUTATION_THRESHOLD */

#if 0
#define MUTATION_THRESHOLD		(VISIBLE_W)
#elif 0
#define MUTATION_THRESHOLD		(VISIBLE_W + VISIBLE_H)	/* more aggressive */
#else
#define MUTATION_THRESHOLD		(VISIBLE_W << 1)	/* even more aggressive */
#endif


WORD	edge_t,			/* top row of changing cells,           inclusive */
		edge_b;			/* bottom row of changing cells,        exclusive */
LPWORD	edge_l,			/* leftmost changing cell on each row,  inclusive */
		edge_r,			/* rightmost changing cell on each row, exclusive */
		edge_L,			/* new values for edge_l */
		edge_R;			/* new values for edge_r */

char no_change_found;

#ifdef LIFE_DEBUGMODE
char mutated;			/* debugging flag: mutate_edges() called */
#endif


#define draw_func_header                    \
	WORD ix, iy, left;                      \
	LPBYTE st, sn, sb;                      \
	WORD sty;                               \
	BYTE e;                                 \
	LPBYTE sts = state_this + UNIVERSE_W;   \
	LPBYTE sns = state_next + UNIVERSE_W;   \
	no_change_found = 1;                    \
	changing_cells = 0;

#if 1

#define select_colour_of_new_cell {             \
	if (e == 0x93) {                            \
		sb = st - UNIVERSE_W - 1;               \
		left = 0;                               \
		if (((*sb) & 0x1F) == 0x11) {           \
			sb = st - 1;                        \
			if (((*sb) & 0x1F) == 0x12) {       \
				sb += UNIVERSE_W;               \
				if (((*sb) & 0x1F) == 0x11)     \
					left = 1;                   \
			}                                   \
			else {                              \
				sb = st - UNIVERSE_W;           \
				if (((*sb) & 0x1F) == 0x12) {   \
					sb++;                       \
					if (((*sb) & 0x1F) == 0x11) \
						left = 1;               \
				}                               \
			}                                   \
			sb = st - UNIVERSE_W - 1;           \
		}                                       \
		if (left) { /* blinker */               \
			changing_cells -= 3;                \
			left = 0;                           \
		}                                       \
		else                                    \
			changing_cells++;                   \
		      if ((*sb) >= 0x30) left++;        \
		sb++; if ((*sb) >= 0x30) left++;        \
		sb++; if ((*sb) >= 0x30) left++;        \
		sb += UNIVERSE_W;                       \
		      if ((*sb) >= 0x30) left++;        \
		sb += UNIVERSE_W;                       \
		      if ((*sb) >= 0x30) left++;        \
		sb--; if ((*sb) >= 0x30) left++;        \
		sb--; if ((*sb) >= 0x30) left++;        \
		sb -= UNIVERSE_W;                       \
		      if ((*sb) >= 0x30) left++;        \
		if (left >= 2) {                        \
			e |= 0x20;                          \
			population[1]++;                    \
		}                                       \
		else                                    \
			population[0]++;                    \
	}                                           \
	else {                                      \
		population[(e & 0x20) >> 5]--;          \
		changing_cells++;                       \
	}                                           \
}

#else

#define select_colour_of_new_cell {             \
	if (e == 0x93) {                            \
		edge_b = 0;                             \
		sb = st - UNIVERSE_W - 1;               \
		      if ((*sb) >= 0x30) edge_b++;      \
		sb++; if ((*sb) >= 0x30) edge_b++;      \
		sb++; if ((*sb) >= 0x30) edge_b++;      \
		sb += UNIVERSE_W;                       \
		      if ((*sb) >= 0x30) edge_b++;      \
		sb += UNIVERSE_W;                       \
		      if ((*sb) >= 0x30) edge_b++;      \
		sb--; if ((*sb) >= 0x30) edge_b++;      \
		sb--; if ((*sb) >= 0x30) edge_b++;      \
		sb -= UNIVERSE_W;                       \
		      if ((*sb) >= 0x30) edge_b++;      \
		if (edge_b >= 2) {                      \
			e |= 0x20;                          \
			population[1]++;                    \
		}                                       \
		else                                    \
			population[0]++;                    \
	}                                           \
	else                                        \
		population[(e & 0x20) >> 5]--;          \
	changing_cells++;                           \
}

#endif

#define do_evolution_of_one_cell(check) {       \
	e = evolution[e];                           \
	if (e & 0x80) {                             \
		if (no_change_found) {                  \
			no_change_found = 0;                \
			edge_t = iy;                        \
		}                                       \
		check;                                  \
		edge_b = iy;                            \
	}                                           \
	*sn = e;                                    \
	st++; sn++;                                 \
}

CHAR cycle(VOID);       /* returns -1 on error */
BOOL initialize(VOID);
CHAR health_check(VOID);

DWORD random_number;
DWORD random_seed;
DWORD random_seed_original;
DWORD rnd(VOID);

#define rand(MAX) (rnd() % (MAX))


#endif	/* LIFE_H_INCLUDED */
