/*************************************************************************
 *
 *  Game of Life -- a screen saver for Windows
 *  Copyright (C) 2007, 2013 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  out_win.h: header file for out_win.c
 *
 ************************************************************************/


#ifndef OUT_WIN_H_INCLUDED
#define OUT_WIN_H_INCLUDED


#include <windows.h>
#include <stdio.h>
#include "tilegfx.h"


/* screen and cell dimensions */

WORD wScreenW, wScreenH;

WORD wCellSize;			/* cell size (pixels) */
WORD wTileOffset[16];


/* bitmaps and device contexts */

HDC hdcOutput;

#ifdef LIFE_VIRTUAL_SCREEN
HBITMAP hbmVirtual;
HDC hdcVirtual;
#else
#ifdef LIFE_VIRTUAL_CANVAS
HBITMAP hbmCanvas;
HDC hdcCanvas;
#endif
#endif

HBITMAP hbmCaptured;
HDC hdcCaptured;
RECT Screen;

HDC hdcTiles;
HBITMAP hbmTiles;
HDC hdcMasks;
HBITMAP hbmMasks;


/* redraw queue */

typedef struct redraw_queue_s {
	WORD x, y;						/* pixel coordinates on screen */
	WORD tx, ty;					/* pixel coordinates to tile array */
	struct redraw_queue_s *next;	/* pointer to next item in the queue */
} redraw_queue_t;

redraw_queue_t *redraw_table;


/* functions in out_win.c */

VOID init_redraw_table(VOID);
VOID destroy_redraw_table(VOID);

VOID (*redraw)(VOID);
VOID redraw_really(VOID);
VOID redraw_not_really(VOID);
VOID (*add_tile_to_redraw_queue)(DWORD, WORD, WORD);
VOID add_tile_to_redraw_queue_really(DWORD, WORD, WORD);
VOID add_tile_to_redraw_queue_not_really(DWORD, WORD, WORD);

VOID (*draw_win_tile_func)(WORD, WORD, WORD, WORD);
VOID (*draw_win_tile_nobg_func)(WORD, WORD, WORD, WORD);
VOID draw_win_tile(WORD, WORD, WORD, WORD);
VOID draw_win_tile_transparent(WORD, WORD, WORD, WORD);
VOID draw_win_tile_transparent_nobg(WORD, WORD, WORD, WORD);

BOOL draw_win_only(LPBYTE);


#endif	/* OUT_WIN_H_INCLUDED */
