/*************************************************************************
 *
 *  Game of Life -- a screen saver for Windows
 *  Copyright (C) 2007, 2013 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  scrlife.h: header file for scrlife.c
 *
 ************************************************************************/


#ifndef SCRLIFE_H_INCLUDED
#define SCRLIFE_H_INCLUDED


#include <windows.h>
#include <scrnsave.h>
#include <stdio.h>


/* version info */

#define LIFE_VERSION		"2.0.2"

#ifdef BUILD64
#define LIFE_BUILD_SIZE_STR	"64-bit"
#else
#define LIFE_BUILD_SIZE_STR	"32-bit"
#endif

#ifdef UNICODE
#define LIFE_BUILD_STR		LIFE_BUILD_SIZE_STR "/Unicode"
#else
#define LIFE_BUILD_STR		LIFE_BUILD_SIZE_STR
#endif

#define AUTHOR_EMAIL		"jussi.hukkanen@gmail.com"


/* tilesets */

#define TILESET_CLASSIC		1
#define TILESET_BLOB		2
#define TILESET_BRICK		3
#define TILESET_PIXEL		4

#define FALLBACK_TILESET	TILESET_PIXEL

#define TILESET_QUARTERS	(wTileStyle != TILESET_PIXEL)

#define MAX_TILESET_SIZE			64

typedef struct tileset_def_s {
	WORD wTileset;	/* one of the TILESET_* macros */
	WORD wSize;		/* value for wCellSize */
	WORD wTileID;	/* resource id from tilegfx.h */
	WORD wMaskID;	/* -"- */
} tileset_def_t;


/* user-configurable options */

LONG lDelay;					/* redraw delay */

COLORREF crA, crB, cr0;

#define COLOUR_A_ENABLED 0x01
#define COLOUR_B_ENABLED 0x02
#define COLOUR_0_ENABLED 0x04
#define OPAQUE_MODE (bColourFlags & COLOUR_0_ENABLED)
#define SINGLE_COLOUR_MODE \
	((bColourFlags & (COLOUR_A_ENABLED | COLOUR_B_ENABLED)) != \
	(COLOUR_A_ENABLED | COLOUR_B_ENABLED))

BYTE bColourFlags;

WORD wTileStyle;				/* one of the TILESET_* macros */

DWORD dwGeneration;				/* starting generation */


/* other */

TCHAR szTemp[200];				/* temporary array of characters */


/* functions in config.c */

BOOL read_config_file(VOID);
BOOL save_config_file(VOID);


#endif	/* SCRLIFE_H_INCLUDED */
